package ColorDemo::Intro;
use v5.38;
use strict;

use Carp;

use Time::HiRes qw(time);

sub new($proto, %config) {
    my $class = ref($proto) || $proto;

    my $self = bless \%config, $class;

    foreach my $name (qw[screen font sound]) {
        if(!defined($self->{$name})) {
            croak("Intro.pm requires $name");
        }
    }

    my $screen = $self->{screen};
    ($self->{screenw}, $self->{screenh}) = $screen->getSize();

    ($self->{logo}, $self->{logow}, $self->{logoh}) = $screen->loadImage('gfx/ball.png');
    ($self->{megademo}, $self->{megademow}, $self->{megademoh}) = $screen->loadImage('gfx/megademo.png');

    $self->{fxnow} = $self->{sound}->loadfx('snd/and_now.wav');
    $self->{fxcavac} = $self->{sound}->loadfx('snd/cavac_presents.wav');
    $self->{fxnisa} = $self->{sound}->loadfx('snd/nisa_design.wav');
    $self->{fxcolor} = $self->{sound}->loadfx('snd/color_demo.wav');

    return $self;
}

sub run($self) {
    my $col = 255;
    my $screen = $self->{screen};
    my $font = $self->{font};
    my $fade = 0.0;

    $screen->setStatus("Intro");

    my $lasttime = time;

    # White flash at the start
    while(0) {
        $screen->clearBuffer(int($col), int($col), int($col));
        my $key = $screen->render();
        if($key) {
            return $key;
        }
        my $now = time;
        my $duration = $now - $lasttime;
        $lasttime = $now;
        if(!$col) {
            last;
        }

        my $stepsize = 255 / (1 / $duration);
        $col -= $stepsize;
        if($col < 0) {
            $col = 0;
        }
    }

    my $x = int(($self->{screenw} - $self->{logow}) / 2);
    my $y = int(($self->{screenh} - $self->{logoh}) / 2);

    $self->{sound}->playfx(2, $self->{fxnow}, 0);
    $screen->clearBuffer(0, 0, 0);

    # Fade in NISA logo
    while(1) {
        $screen->setImage($x, $y, $self->{logo}, $fade);
        my $key = $screen->render();
        if($key) {
            return $key;
        }
        my $now = time;
        my $duration = $now - $lasttime;
        $lasttime = $now;
        if($fade >= 1.0) {
            last;
        }
        my $stepsize = 1.0 / (1 / $duration);
        $fade+= $stepsize;
    }

    # Fade out NISA logo
    while(1) {
        $screen->setImage($x, $y, $self->{logo}, $fade);
        my $key = $screen->render();
        if($key) {
            return $key;
        }
        my $now = time;
        my $duration = $now - $lasttime;
        $lasttime = $now;
        if($fade <= 0.0) {
            last;
        }
        my $stepsize = 1.0 / (1 / $duration);
        $fade-= $stepsize;
    }


    my @textscreens = (
        ['cavac presents', 255, 255, 255, 2, 'fxcavac'],
        ["a\nNULL ISLAND SPACE AGENCY\ndesign", 0, 255, 0, 3, 'fxnisa'],
    );

    foreach my $page (@textscreens) {
        $screen->clearBuffer(0, 0, 0);
        $fade = 0;
        my ($text, $r, $g, $b, $fadetime, $fxname) = @{$page};
        $self->{sound}->playfx(2, $self->{$fxname}, 0);

        # Fade in
        while(1) {
            $font->renderCenterText($text, $r * $fade, $g * $fade, $b * $fade);
            my $key = $screen->render();
            if($key) { 
                return $key;
            }
            my $now = time;
            my $duration = $now - $lasttime;
            $lasttime = $now;
            if($fade >= 1.0) {
                last;
            }
            my $stepsize = 1.0 / (1 / $duration);
            $fade+= $stepsize;
        }

        # Fade out
        $fade = 1.0;
        while(1) {
            $font->renderCenterText($text, $r * $fade, $g * $fade, $b * $fade);
            my $key = $screen->render();
            if($key) {
                return $key;
            }
            my $now = time;
            my $duration = $now - $lasttime;
            $lasttime = $now;
            if($fade <= 0.0) {
                last;
            }
            my $stepsize = 1.0 / ($fadetime / $duration);
            $fade-= $stepsize;
        }
    }
    
    $x = int(($self->{screenw} - $self->{megademow}) / 2);
    $y = int(($self->{screenh} - $self->{megademoh}) / 2);
    $fade = 0.0;

    $self->{sound}->playfx(2, $self->{fxcolor}, 0);
    $screen->clearBuffer(0, 0, 0);
    # Fade in MEGADEMO logo
    while(1) {
        $screen->setImage(0, 0, $self->{megademo}, $fade);
        my $key = $screen->render();
        if($key) {
            return $key;
        }
        my $now = time;
        my $duration = $now - $lasttime;
        $lasttime = $now;
        if($fade >= 1.0) {
            last;
        }
        my $stepsize = 1.0 / (1 / $duration);
        $fade+= $stepsize;
    }

    $lasttime = time;
    while(1) {
        if(($lasttime + 3.5) < time) {
            last;
        }
        my $key = $screen->render();
        if($key) {
            return $key;
        }
    }

    # Fade out MEGADEMO logo
    $fade = 1.0;
    $lasttime = time;
    while(1) {
        $screen->setImage(0, 0, $self->{megademo}, $fade);
        my $key = $screen->render();
        if($key) {
            return $key;
        }
        my $now = time;
        my $duration = $now - $lasttime;
        $lasttime = $now;
        if($fade <= 0.0) {
            last;
        }
        my $stepsize = 1.0 / (1 / $duration);
        $fade-= $stepsize;
    }

    return 0;
}
