package ColorDemo::Screen;
use v5.38;
use strict;

use Term::ANSIScreen qw/:cursor :screen/;
use GD;
use Carp;

use Term::ReadKey qw(GetTerminalSize ReadKey ReadMode);
use Time::HiRes qw(time);

my $halfblock = chr(9600);
utf8::encode($halfblock);

my $fullblock = chr(9608);
utf8::encode($fullblock);

my $darkshade = chr(9619);
utf8::encode($darkshade);

my $lightshade = chr(9617);
utf8::encode($lightshade);

sub new($proto, %config) {
    STDOUT->autoflush(1);

    my $class = ref($proto) || $proto;

    my $self = bless \%config, $class;

    my ($cols, $rows, $wpixels, $hpixels) = GetTerminalSize();

    if($cols < 270 || $rows < 60) {
        croak("Your terminal needs to be at least 270x60 in size, but we have only $cols x $rows");
    }

    if($cols > 270) {
        $cols = 270;
    }

    if($rows > 60) {
        $rows = 60;
    }

    # Make sure we don't do accidental line wraps or scrolls
    $cols--;
    $rows--;

    # Last line is reserved for status information
    $rows--;

    # We can do two pixels per row
    $rows *= 2;

    $self->{rows} = $rows;
    $self->{cols} = $cols;

    print "$cols x $rows\n";

    my @img;
    for(my $i = 0; $i < ($cols * $rows); $i++) {
        my @tmp = (0, 0, 0);
        push @img, \@tmp;
    }

    $self->{img} = \@img;

    $self->{statustext} = "";

    $self->{lastrender} = 0;
    $self->{progress} = -1;
    $self->{info} = '';

    $self->{cls} = cls();
    $self->{home} = locate();

    $self->clear();

    ReadMode 4;

    return $self;
}

sub reset($self) {
    ReadMode 0;
    $self->clear();
}

sub getSize($self) {
    return ($self->{cols}, $self->{rows});
}

sub clear($self) {
    #syswrite(STDOUT, $self->{cls});
    print $self->{cls};
    return;
}

sub setStatus($self, $val) {
    $val .= ' ' x (30 - length($val));
    $self->{statustext} = $val;
    return;
}

sub setProgress($self, $progress) {
    $self->{progress} = $progress;
    return;
}

sub setInfo($self, $info) {
    $self->{info} = $info;
    return;
}

sub setPixel($self, $x, $y, $r, $g, $b) {
    if($x >= $self->{cols} || $y >= $self->{rows} || $x < 0 || $y < 0) {
        return;
    }

    $r = int($r);
    $g = int($g);
    $b = int($b);

    if($r < 0) {
        $r = 0;
    } elsif($r > 255) {
        $r = 255;
    }

    if($g < 0) {
        $g = 0;
    } elsif($g > 255) {
        $g = 255;
    }

    if($b < 0) {
        $b = 0;
    } elsif($b > 255) {
        $b = 255;
    }

    my $idx = $x + ($y * $self->{cols});
    $self->{img}->[$idx]->[0] = $r;
    $self->{img}->[$idx]->[1] = $g;
    $self->{img}->[$idx]->[2] = $b;

    return;
}

sub clearBuffer($self, $r, $g, $b) {
    $r = int($r);
    $g = int($g);
    $b = int($b);

    if($r < 0) {
        $r = 0;
    } elsif($r > 255) {
        $r = 255;
    }

    if($g < 0) {
        $g = 0;
    } elsif($g > 255) {
        $g = 255;
    }

    if($b < 0) {
        $b = 0;
    } elsif($b > 255) {
        $b = 255;
    }

    for(my $idx = 0; $idx < scalar @{$self->{img}}; $idx++) {
        $self->{img}->[$idx]->[0] = $r;
        $self->{img}->[$idx]->[1] = $g;
        $self->{img}->[$idx]->[2] = $b;
    }

    return;
}


sub loadImage($self, $fname) {
    my $img = GD::Image->new($fname);
    my ($origw, $origh) = $img->getBounds();
    my ($w, $h) = ($origw + 0, $origh + 0);

    my $divfactor = 1;
    if($w > $self->{cols}) {
        my $tmp = $w / $self->{cols};
        #print "$w / $self->{cols} / $tmp\n";
        if($tmp > $divfactor) {
            $divfactor = $tmp;
        }
    }
    if($h > $self->{rows}) {
        my $tmp = $h / $self->{rows};
        #print "$h / $self->{rows} / $tmp\n";
        if($tmp > $divfactor) {
            $divfactor = $tmp;
        }
    }

    if($divfactor > 1) {
        croak("DIVFACTOR $divfactor");
        $w = int($w / $divfactor);
        $h = int($h / $divfactor);

        my $newpic = GD::Image->new($w, $h, 1);

        $newpic->copyResized($img,
                          0, 0, # DEST X Y
                          0, 0, # SRC X Y
                          $w, $h, # DEST W H
                          $origw, $origh, # SRC W H
                        );

        $img = $newpic;
    }

    return ($img, $w, $h);
}

sub setImage($self, $x, $y, $img, $fade = 1.0) {
    my ($w, $h) = $img->getBounds();

    $x = int($x);
    $y = int($y);

    if($fade < 0.0) {
        $fade = 0.0;
    } elsif($fade > 1.0) {
        $fade = 1.0;
    }

    # Crop image if required
    if(1) {
        $w += $x;
        if($w > $self->{cols}) {
            $w = $self->{cols};
        }
        $w -= $x;

        $h += $y;
        if($h > $self->{rows}) {
            $h = $self->{rows};
        }
        $h -= $y;
    }

    my ($r, $g, $b);
    for(my $xidx = 0; $xidx < $w; $xidx++) {
        for(my $yidx = 0; $yidx < $h; $yidx++) {
            my $color = $img->getPixel($xidx, $yidx);
            ($r,$g,$b) = $img->rgb($color);

            $r = ($r * $fade);
            $g = ($g * $fade);
            $b = ($b * $fade);

            my $imgidx = ($xidx + $x) + (($yidx + $y) * $self->{cols});
            $self->{img}->[$imgidx]->[0] = int($r);
            $self->{img}->[$imgidx]->[1] = int($g);
            $self->{img}->[$imgidx]->[2] = int($b);
        }
    }

    return;
}

sub crossFadeImage($self, $x, $y, $oldimg, $newimg, $fade = 1.0) {
    my ($w, $h) = $newimg->getBounds();

    $x = int($x);
    $y = int($y);

    if($fade < 0.0) {
        $fade = 0.0;
    } elsif($fade > 1.0) {
        $fade = 1.0;
    }

    # Crop image if required
    if(1) {
        $w += $x;
        if($w > $self->{cols}) {
            $w = $self->{cols};
        }
        $w -= $x;

        $h += $y;
        if($h > $self->{rows}) {
            $h = $self->{rows};
        }
        $h -= $y;
    }

    my ($r, $g, $b);
    my ($newr, $newg, $newb);
    my ($oldr, $oldg, $oldb);
    for(my $xidx = 0; $xidx < $w; $xidx++) {
        for(my $yidx = 0; $yidx < $h; $yidx++) {
            my $newcolor = $newimg->getPixel($xidx, $yidx);
            ($newr,$newg,$newb) = $newimg->rgb($newcolor);

            my $oldcolor = $oldimg->getPixel($xidx, $yidx);
            ($oldr,$oldg,$oldb) = $oldimg->rgb($oldcolor);

            $r = ($newr * $fade) + ($oldr * (1.0 - $fade));
            $g = ($newg * $fade) + ($oldg * (1.0 - $fade));
            $b = ($newb * $fade) + ($oldb * (1.0 - $fade));

            my $imgidx = ($xidx + $x) + (($yidx + $y) * $self->{cols});
            $self->{img}->[$imgidx]->[0] = int($r);
            $self->{img}->[$imgidx]->[1] = int($g);
            $self->{img}->[$imgidx]->[2] = int($b);
        }
    }

    return;
}

sub render($self) {
    my $lastfgcolor = '';
    my $lastbgcolor = '';

    my $out = '' . $self->{home};
    my ($r, $g, $b); # Color vars
    for(my $y = 0; $y < $self->{rows}; $y+=2) {
        for(my $x = 0; $x < $self->{cols}; $x++) {
            # Foreground color
            ($r,$g,$b) = @{$self->{img}->[$x + ($y * $self->{cols})]};

            my $newfgcolor = "\e[38;2;" . join(';', $r, $g, $b) . "m";

            if($newfgcolor ne $lastfgcolor) {
                $lastfgcolor = $newfgcolor;
                $out .= $newfgcolor;
            }

            # Background color
            my $lowy = $y + 1;
            if($lowy == $self->{rows}) {
                # End of image. need a black half-line
                ($r, $g, $b) = (0, 0, 0);
            } else {
                ($r,$g,$b) = @{$self->{img}->[$x + ($lowy * $self->{cols})]};
            }

            my $newbgcolor = "\e[48;2;" . join(';', $r, $g, $b) . "m";

            if($newbgcolor ne $lastbgcolor) {
                $lastbgcolor = $newbgcolor;
                $out .= $newbgcolor;
            }


            $out .= $halfblock;
            #print utf8::encode("\N{FULL BLOCK}");
        }
        ($r, $g, $b) = (0, 0, 0);
        $lastfgcolor = "\e[38;2;" . join(';', 255, 255, 255) . "m";
        $lastbgcolor = "\e[48;2;" . join(';', 0, 0, 0) . "m";
        $out .= $lastfgcolor . $lastbgcolor . "\n";
    }

    my $now = time;
    my $fps = 0;
    if($self->{lastrender}) {
        $fps = (int((1 / ($now - $self->{lastrender})) * 100)/100);
    }
    $self->{lastrender} = $now;

    # Status line
    my ($pre, $post) = split/\./o, $fps;
    if(!defined($post)) {
        $post = '00';
    }
    while(length($pre) < 4) {
        $pre = ' ' . $pre;
    }
    while(length($post) < 2) {
        $post .= '0';
    }

    $fps = 'FPS: ' . $pre . '.' . $post;
        
    $fps .= ' ' x (18 - length($fps));
    $out .= $fps;

    $out .= $self->{statustext};

    if($self->{info} ne '') {
        $out .= $self->{info};
        $out .= ' ' x (220 - length($self->{info}));
    } elsif($self->{progress} == -1) {
        $out .= ' ' x 220;
    } elsif($self->{info} ne '') {
        $out .= $self->{info};
        $out .= ' ' x (220 - length($self->{info}));
    } else {
        $out .= $fullblock;
        my $full = int(218 * $self->{progress});
        $out .= $darkshade x $full;
        $out .= $lightshade x (218-$full);
        $out .= $fullblock;
    }

    print $out;

    my $key = ReadKey(-1);

    if(defined($key)) {
        $key = ord($key);
        #croak("\n\n\n ** $key **\n");
        if($key == 32) {
            return 1;
        } elsif($key == 27 || $key == 113) {
            return -1;
        }
    }

    return 0;
}


